"""extract information from html source code of bilibili.com."""
from urllib.parse import urljoin
import json
import re

from video_dl.extractor import Extractor


class BilibiliVideoExtractor(Extractor):
    """bilibili information extractor."""
    pattern = [
        re.compile('bilibili.com/video/BV.*'),
    ]

    # re patterns to extract information from html source code
    re_state = re.compile(r'__INITIAL_STATE__=(.*?);\(function\(\)')
    re_playinfo = re.compile(r'window.__playinfo__=(.*?)</script>')

    def __init__(self):
        self.id2desc = None

    def get_title(self, resp: str) -> str:
        """get video's title from html source code."""
        state = json.loads(self.re_state.search(resp).group(1))

        current_page = state['p']
        pages = state['videoData']['pages']

        if len(pages) == 1:
            return state['videoData']['title']

        for page in pages:
            if page['page'] == current_page:
                return page['part']

    def get_parent_folder(self, resp: str) -> str:
        """get video's parent folder from html source code."""
        state = json.loads(self.re_state.search(resp).group(1))

        pages = state['videoData']['pages']
        if len(pages) > 1:
            return state['videoData']['title']
        else:
            return None

    def get_pictures(self, resp: str) -> list:
        """get pictures' information from html source code."""
        playinfo = json.loads(self.re_playinfo.search(resp).group(1))

        if self.id2desc is None:
            desc = playinfo['data']['accept_description']
            quality = playinfo['data']['accept_quality']
            self.id2desc = {
                str(key): value for key, value in zip(quality, desc)
            }

        pictures = playinfo['data']['dash']['video']
        for media in pictures:
            yield {
                'url': media['base_url'],
                'size': media['bandwidth'],
                'desc': f"{self.id2desc[str(media['id'])]} + {media['codecs']}"
            }

    def get_sounds(self, resp: str) -> list:
        """get sounds' information from html source code."""
        playinfo = json.loads(self.re_playinfo.search(resp).group(1))

        if self.id2desc is None:
            desc = playinfo['data']['accept_description']
            quality = playinfo['data']['accept_quality']
            self.id2desc = {
                str(key): value for key, value in zip(quality, desc)
            }

        sounds = playinfo['data']['dash']['audio']
        for media in sounds:
            yield {
                'url': media['base_url'],
                'size': media['bandwidth'],
            }

    def generate_urls(self, resp: str, base_url: str) -> list:
        """generate urls from html resource code."""
        state = json.loads(self.re_state.search(resp).group(1))

        current_page = state['p']
        pages = state['videoData']['pages']
        for page in pages:
            if (index := page['page']) != current_page:
                yield urljoin(base_url, f'?p={index}')


class BilibiliBangumiExtractor(BilibiliVideoExtractor, Extractor):
    """extractor for bilibili bangumi"""
    pattern = [
        re.compile('bilibili.com/bangumi/play/ep.*'),
        re.compile('bilibili.com/bangumi/play/ss.*'),
    ]

    def get_title(self, resp: str) -> str:
        """get video's title from html source code."""
        state = json.loads(self.re_state.search(resp).group(1))
        return state['h1Title']

    def get_parent_folder(self, resp: str) -> str:
        """get video's parent folder from html source code."""
        state = json.loads(self.re_state.search(resp).group(1))
        return state['mediaInfo']['season_title']

    def generate_urls(self, resp: str, base_url: str) -> list:
        """generate urls from html resource code."""
        state = json.loads(self.re_state.search(resp).group(1))

        pages = state['mediaInfo']['episodes']
        for page in pages:
            url = page['link'].replace('/u002f', '/')
            if url not in base_url:
                yield url



"""toolbox which provides lots of helper function.

class UserAgent deals with user agent.

Avaliable function:
    UserAgent().random: get a random user agent.
    info: print prompt message.
    ask_user: ask user to provide a string of numbers.

Typical usage:
    random_ua = UserAgent().random
    info('url', 'https://www.bilibili.com/video/xxx')
    video_index, audio_index = ask_user(count=2, default=1)
"""
from typing import Optional
import os
import random


class UserAgent(object):
    """user agent.

    Attributes:
        random: get a random user agent.
    """

    def __init__(self, file_path: str = None):
        """read user agent list from a txt file.

        Args:
            file_path: txt file path, default: ./resource/user_agents.txt.
        """
        if file_path is None:
            file_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'resource', 'user_agents.txt'
            )

        with open(file_path, 'r', encoding='utf-8') as f:
            self.user_agents = [line.strip()
                                for line in f.readlines() if line.strip()]

    @property
    def random(self) -> str:
        """return a random user agent"""
        return random.choice(self.user_agents)


class ConsoleColor(object):
    """ANSI control code for color."""
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def info(label: str, *args, **kwargs) -> None:
    """print information to console with colors."""
    print(f'{ConsoleColor.WARNING}[{label}]{ConsoleColor.OKGREEN}',
          *args, ConsoleColor.ENDC, **kwargs)


def ask_user(count: Optional[int] = 2, default: Optional[int] = 1) -> tuple:
    """ask user to provide answers.

    Args:
        count: count of required answer.
        default: default value.
    """
    prompt = ('What is your answer(space to separate, enter to use default: '
              f'{default}):')
    answer = input(prompt).strip()
    if not answer:
        return (default, ) * count
    else:
        return tuple(int(item) for item in answer.split(' '))



"""read config from config file and user's input, then return the result.

Typical usage:
    args = Arguments()
    url = args.url
    cookie = args.cookie
    proxy = args.proxy
    interactive_flag = args.interactive

Available arguments:
    interactive: choose media resource manually.
    lists: try to find a playlist and download all video contained in it.
    directory: set a target directory to save video.
    cookie: user's own cookie.
    proxy: internet proxy.
    url: target url.
    max_conn: maximum connections simultaneously.
    chunk_size: chunk size to read from stream.
    big_file_threshold: file size exceeds this threshold will be sliced.
"""
from typing import Any
import argparse
import json
import os

from video_dl import __version__ as version


class Config(object):
    """read, parse and return config information."""

    def __init__(self, file_path: str = None):
        """read config information from a json file.

        Args:
            file_path: json file path, default: resource/config.json
        """
        # set config file's absolute path
        if file_path is None:
            file_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'resource', 'config.json'
            )

        with open(file_path, 'r', encoding='utf-8') as f:
            self.config = json.load(f)


class ArgParse(object):
    """parse command line arguments with argparse."""
    def __init__(self):
        parser = argparse.ArgumentParser(
            prog='video_dl',
            description='A naive online video downloader based on aiohttp',
            epilog=('You could find more important information in '
                    '[github](https://github.com/fengdongfa1995/video_dl).'),
        )

        # flags
        parser.add_argument(
            '-i', '--interactive', action='store_true',
            help='Manually select download resources.',
        )

        parser.add_argument(
            '-l', '--lists', action='store_true',
            help='try to find a playlist and download all videos in it.',
        )

        # something provided by user
        parser.add_argument(
            '-d', '--directory',
            help='set target diretory to save video file(s).',
        )

        parser.add_argument(
            '-c', '--cookie',
            help='provide your cookie.',
        )

        parser.add_argument(
            '-p', '--proxy',
            help='set proxy. e.g.: http://127.0.0.1:10809',
        )

        # required arguments
        parser.add_argument(
            'url', help='target url copied from online video website.',
        )

        # program's version
        parser.add_argument('-v', '--version',
                            action='version', version=f'%(prog)s {version}')

        # command line arguments parser result
        args = parser.parse_args()
        self.args = vars(args)


class Arguments(object):
    """provide global variables to other modules."""
    args = ArgParse().args  # arguments provided by user
    config = Config().config  # auguments provided by config file

    def _if_none_return_empty_string(self, key: str) -> str:
        """if value of the key is None, the return ''."""
        # just find key in self.args
        # because config file have no key point to None
        return '' if self.args[key] is None else self.args[key]

    @property
    def cookie(self) -> str:
        """
        if user doesn't set a cookie, cookie will be None.
        but our program need a empty string not the None.
        """
        return self._if_none_return_empty_string('cookie')

    def __getattr__(self, key: str) -> Any:
        if key not in self.args and key not in self.config:
            raise KeyError
        elif key not in self.args and key in self.config:
            value = self.config[key]
        elif key in self.args and key not in self.config:
            value = self.args[key]
        elif key in self.args and key in self.config:
            if self.args[key] is None:
                value = self.config[key]
            else:
                value = self.args[key]

        setattr(self, key, value)
        return value



"""Base class of all Spiders.

Typical usage:
    url = 'https://www.bilibili.com/video/BV15L411p7M8'
    spider = Spider.create_spider(url)  # will return a BilibiliSpider object
    asycio.run(spider.run())  # try to fetch resource information and download
"""
from urllib.parse import urlparse
import aiohttp
import asyncio

from video_dl.args import Arguments
from video_dl.toolbox import UserAgent, info


class Spider(object):
    """crawl data from web and download video.

    subclass of Spider should provide two public attributes:
        site: this class will use this field to create a specific Spider for
            target url. for example, BilibiliSpider's site is 'bilibili.com',
            will auto match to a url like 'https://www.bilibili.com/video/*'.
        home_url: target website's home page. this field will be inserted into
            headers of session to avoid some `no referer, no download` policy.

    subclass of Spider should implement some public methods:
        before_download: do something before download, just like: parse html.
        after_download: merge picture and sound to a completed video, delete
            tamporary files, and et al..
    """
    arg = Arguments()

    cookie = arg.cookie
    diretory = arg.directory
    proxy = arg.proxy
    url = arg.url
    lists = arg.lists

    @classmethod
    def create(cls, url: str):
        """create a specific subclass depends on url."""
        netloc = urlparse(url).netloc
        for subclass in cls.__subclasses__():
            if subclass.site in netloc:
                return subclass()
        raise NotImplementedError

    def __init__(self):
        self.session = None
        self.headers = {
            'accept': '*/*',
            'user-agent': UserAgent().random,
            'cookie': self.cookie,
            'referer': self.home_url,
            'origin': self.home_url,
        }

        # list that contains Videos
        self.video_list = []

    async def create_session(self) -> None:
        """create client seesion if not exist."""
        if self.session is None:
            conn = aiohttp.connector.TCPConnector(
                force_close=True, enable_cleanup_closed=True, verify_ssl=False
            )
            self.session = aiohttp.ClientSession(headers=self.headers,
                                                 connector=conn)

    async def close_session(self) -> None:
        """close client session if possible."""
        if self.session:
            await self.session.close()

    async def fetch_html(self, url: str) -> tuple:
        """get url's html source code from internet."""
        async with self.session.get(url=url, proxy=self.proxy) as r:
            # maybe exist redirection
            # TODO: watch out more redirections to modify index in r.history
            if r.history:
                url = r.history[0].headers['location']
            return await r.text(), url

    async def before_download(self) -> None:
        """do something before download"""
        raise NotImplementedError

    async def downloading(self) -> None:
        """download video from web."""
        tasks = []
        for video in self.video_list:
            video.choose_collection()
            tasks.append(video.download())

        await asyncio.gather(*tasks)

    def after_downloaded(self) -> None:
        """do something after downloaded video."""
        pass

    async def run(self) -> None:
        """start crawl."""
        info('site', self.site)
        await self.create_session()

        await self.before_download()
        await self.downloading()
        self.after_downloaded()

        await self.close_session()



"""Library for handling video.

Some online video website stores their video (picture) and audio (sound)
separately in differrent place. So, if we want to download a complete video,
we should download video and audio (I say, media) first, and then merge them to
one.

when downloading a single media, if its size exceeds a certain threshold,
program will slice it to many fragments and download with different coroutine.

Available function:
    - Media().download: download a media from internet.
    - MediaCollection().download: download medias contained in MediaCollection.
    - MediaCollection().merge: combined medias contained in collection to one.

Typical usage example:
    media1 = Media(**{'url': 'url1', size: '1', desc: '.'})
    media2 = Media(**{'url': 'url2', size: '2', desc: '.'})
    media_collection = MediaCollation([media1, media2])
    media_collection.download()
    media_collection.merge()  # if necessary
"""
from typing import List, Optional
import aiohttp
import asyncio
import contextvars
import math
import os
import subprocess

from prettytable import PrettyTable

from video_dl.args import Arguments
from video_dl.toolbox import ConsoleColor, info, ask_user


session = contextvars.ContextVar('Aiohttp.ClientSession', default=None)
semaphore = contextvars.ContextVar('asyncio.Semaphore', default=None)


class Media(object):
    """Class used to handle media."""
    arg = Arguments()

    _chunk_size = arg.chunk_size
    _threshold = arg.big_file_threshold

    def __init__(self, *, url: str,
                 size: Optional[int] = 0,
                 desc: Optional[str] = 'null'):
        """Initialize a media object.

        Args:
            url: target url.
            size: media's file size. will be used to sort.
            desc: description of media, default: null.
        """
        self.url = url  # download media from this url
        self.size = size  # file size fetched from server, will be used to sort
        self.desc = desc  # description for choosing by user

        # download to this location, will be changed by MediaCollection outside
        self.location = None

        # file size during downloading, will be used to draw a progress bar.
        self._current_size = 0

    def _get_location(self, index: Optional[int] = 0) -> str:
        """get media slice's target storage path.

        insert a index into file's name and return.

        Args:
            index: index of media slice. defalut value is 0 means no slice.

        Returns:
            media slice's target storage location. e.g.: ./media_p2.mp4.
        """
        if index == 0:
            return self.location
        else:
            folder, name = os.path.split(self.location)
            title, suffix = os.path.splitext(name)
            return os.path.join(folder, f'{title}_p{index}{suffix}')

    async def _set_size(self) -> None:
        """set media file's real size by parsing server's response headers."""
        headers = {'range': 'bytes=0-1'}
        async with session.get().get(url=self.url, headers=headers) as r:
            self.size = int(r.headers['Content-Range'].split('/')[1])

    def _get_headers(self, index: Optional[int] = 0) -> dict:
        """get a headers should be sent to server for downloading a media slice

        Args:
            index: index of media slice. begin with 1.
                default value 0 means no slice.

        Returns:
            a request header with slice range. defalut: {}.
        """
        if index == 0:
            return {}

        slice_point = range(0, self.size, self._threshold)
        start_point = slice_point[index - 1]
        end_point = min(start_point + self._threshold - 1, self.size - 1)
        return {'range': f'bytes={start_point}-{end_point}'}

    async def download(self) -> None:
        """download media to target location."""
        info('ready to download', os.path.split(self.location)[1])

        await self._set_size()
        if self.size <= self._threshold:  # don't need silce
            await self._download_slice()
            print()  # avoid overwritten
        else:
            # slice media, create task and run
            slice_count = math.ceil(self.size / self._threshold)
            await asyncio.wait([
                asyncio.create_task(self._download_slice(index + 1))
                for index in range(slice_count)
            ])

            print()  # avoid overwritten
            info('slices2one', f'merging to {os.path.split(self.location)[1]}')
            with open(self.location, 'wb') as f:
                for index in range(slice_count):
                    target = self._get_location(index + 1)
                    with open(target, 'rb') as media_slice:
                        f.write(media_slice.read())
                    os.remove(target)

    async def _download_slice(self, index: Optional[int] = 0) -> None:
        """download media slice from internet.

        @param index: index of media slice which we want to download.
                    default value 0 means no slice.
        """
        target = self._get_location(index)  # get target location to save media
        headers = self._get_headers(index)  # get headers to send to server

        async with semaphore.get():
            async with session.get().get(url=self.url, headers=headers) as r:
                with open(target, 'wb') as f:
                    async for chk in r.content.iter_chunked(self._chunk_size):
                        f.write(chk)

                        self._current_size += self._chunk_size
                        self._print_progress()

    def _print_progress(self) -> None:
        """print a naive progress bar."""
        progress = int(self._current_size / self.size * 20)
        print('\r', ConsoleColor.WARNING, '[downloading] ',
              ConsoleColor.OKGREEN,
              f'[{progress*100/20:3.0f}%]',  # percent
              f'({self._current_size/1024/1024:6.2f}/',  # current_size
              f'{self.size/1024/1024:6.2f}MB)|',  # total_size
              'x' * progress, '.' * (20 - progress),  # naive progress bar
              ' ', ConsoleColor.OKCYAN, os.path.split(self.location)[1],
              ConsoleColor.ENDC, '\r', ConsoleColor.ENDC, sep='', end='')


class MediaCollection(list):
    """class for handle list of medias."""
    arg = Arguments()

    def __init__(self, members: List[Media] = None, *,
                 salt: Optional[str] = ''):
        """Initialization

        Args:
            members: list of Medias.
            salt: a string will be inserted into file name.
        """
        if members is None:
            members = []
        super().__init__(members)

        self.salt = salt
        self.location = None  # will be set by Video outsite.

    def get_location(self) -> str:
        """return media collection's location, add salt."""
        if self.salt == '':
            return self.location
        else:
            folder, name = os.path.split(self.location)
            title, suffix = os.path.splitext(name)
            return os.path.join(folder, f'{title}_{self.salt}{suffix}')

    def add_media(self, media: Media) -> None:
        """add media resource into media collection."""
        if media.location is None:
            media.location = self.get_location()
        super().append(media)

    async def download(self) -> None:
        """download all medias contained in this collection."""
        await asyncio.wait([
            asyncio.create_task(item.download()) for item in self
        ])

    def __str__(self):
        """print this media collection with pretty format."""
        tb = PrettyTable()
        tb.field_names = ['index', 'name', 'desc', 'size']
        for index, media in enumerate(self, 1):
            tb.add_row([
                index, os.path.split(media.location)[1], media.desc, media.size
            ])
        return tb.get_string()

    def merge(self) -> None:
        """merge all medias into a complete one.

        only workable after all medias are ready.
        """
        info('merge', f'merging to {self.location} ...')

        # command line command
        cmd = ['ffmpeg']
        for item in self:
            cmd += ['-i', item.location]
        cmd += ['-codec', 'copy', self.location, '-y']

        # call command provided by opration system
        subprocess.run(cmd,
                       stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT,
                       check=True)

        # remove temporary files
        for item in self:
            os.remove(item.location)

    def sort_media(self, reverse: bool = True) -> None:
        """sort medias in media collection, biggest one will be the first."""
        super().sort(key=lambda item: item.size, reverse=reverse)


class Video(object):
    """presents a video."""
    arg = Arguments()

    directory = arg.directory
    interactive = arg.interactive
    lists = arg.lists
    max_conn = arg.max_conn

    def __init__(self, client_session: aiohttp.ClientSession,
                 suffix: Optional[str] = 'mp4'):
        """Initialize a video object.

        Args:
            client_session: session used to access web.
            suffix: the suffix of video file. default: mp4.
        """
        if session.get() is None:
            session.set(client_session)

        if semaphore.get() is None:
            semaphore.set(asyncio.Semaphore(self.max_conn))

        # attributes read from config file or user's input
        self.root_folder = self.directory
        self.use_parent_folder = self.lists

        # attributes should be set by spider
        self.suffix = suffix
        self.parent_folder = None
        self.title = None

        # media collection used to hold url and target location
        self.media_collection = {
            'picture': MediaCollection(salt='picture'),  # video without sound
            'sound': MediaCollection(salt='sound'),  # video without picture
            'video': MediaCollection(),  # video = picture + sound
        }

    def get_folder(self) -> str:
        """return video's store folder."""
        if self.parent_folder is not None and self.use_parent_folder is True:
            return os.path.join(self.root_folder, self.parent_folder)
        return self.root_folder

    def get_location(self) -> str:
        """return video's store location."""
        return os.path.join(self.get_folder(), f'{self.title}.{self.suffix}')

    def add_media(self, media: Media, target: Optional[str] = 'video') -> None:
        """add a media to the specific media collection.

        Args:
            media: a Media object represents some information about media.
            target: the collection will add a media resource. default value is
                'video', means the media with picture and sound.
        """
        if self.media_collection[target].location is None:
            self.media_collection[target].location = self.get_location()
        self.media_collection[target].add_media(media)

    def choose_collection(self) -> MediaCollection:
        """choose download task from media collection.

        Args:
            flag: Do you want to choose media by yourself?

        Returns:
            download task
        """
        if len(self.media_collection['video']) != 0:
            self.media_collection['video'].sort_media()

            # choose from 'video' media collection
            raise NotImplementedError('just wait!')
        else:
            # sort item in media collection
            self.media_collection['picture'].sort_media()
            self.media_collection['sound'].sort_media()

            # choose from 'picture' and 'sound' media collection
            if self.interactive is False:
                info('choose', 'using default value (the first one)...')
                v, a = 1, 1
            else:
                for key in ['picture', 'sound']:
                    info('choose', f'please choose a {key} below...')
                    print(self.media_collection[key])
                v, a = ask_user(count=2, default=1)

            info('choosed', '↓↓↓↓↓↓↓↓↓↓↓')
            self.add_media(self.media_collection['picture'][v - 1])
            self.add_media(self.media_collection['sound'][a - 1])
            print(self.media_collection['video'])

    async def download(self) -> None:
        """download medias contained in video media collection."""
        # create directory if necessary
        folder = self.get_folder()
        if folder != self.root_folder:
            if os.path.exists(folder) is False or os.path.isfile(folder):
                os.mkdir(self.get_folder())

        await self.media_collection['video'].download()

    def merge(self) -> None:
        """merge medias contained in video media collection."""
        self.media_collection['video'].merge()



"""Spider for bilibili.com"""
import asyncio
import re

from video_dl.spider import Spider
from video_dl.toolbox import info
from video_dl.video import Video, Media
from video_dl.extractor import Extractor


class BilibiliSpider(Spider):
    """spider for bilibili.com"""
    site = 'bilibili.com'
    home_url = 'https://www.bilibili.com'

    pattern = [
        re.compile('bilibili.com/bangumi/play/ep.*'),
        re.compile('bilibili.com/video/BV.*'),
    ]

    def __init__(self):
        super().__init__()

        # video definition's id to definition's description
        self.id2desc = None
        self.list_video_already_flag = False

    async def before_download(self) -> None:
        await self.parse_html(self.url)

    def after_downloaded(self) -> None:
        for video in self.video_list:
            video.merge()

    async def parse_html(self, target_url: str) -> None:
        """extract key information from html source code.

        Args:
            target_url: target url copied from online vide website.
        """
        info('url', target_url)
        resp, target_url = await self.fetch_html(target_url)
        extractor = Extractor.create(target_url)

        # should we extract video information in this target_url
        extract_flag = False
        for pattern in self.pattern:
            if pattern.search(target_url):
                extract_flag = True
                break

        if extract_flag is False:
            urls = extractor.generate_urls(resp, self.url)
            await self.parse_html(next(urls))
            return

        video = Video(self.session)  # need a session to access internet
        video.title = extractor.get_title(resp)

        if self.lists:
            video.parent_folder = extractor.get_parent_folder(resp)

        try:
            for picture in extractor.get_pictures(resp):
                video.add_media(Media(**picture), target='picture')
        except Exception:  # pylint: disable=W0703
            info(
                'failed',
                f'check authority in {target_url} or post it to github issues.'
            )
            return

        for sound in extractor.get_sounds(resp):
            video.add_media(Media(**sound), target='sound')

        self.video_list.append(video)

        if self.lists and self.list_video_already_flag is False:
            info('list', 'tring to fetch more videos...')
            self.list_video_already_flag = True
            tasks = [self.parse_html(url)
                     for url in extractor.generate_urls(resp, self.url)]

            if tasks:
                info('list', f'fetched {len(tasks)} more video(s)...')
                await asyncio.gather(*tasks)
            else:
                info('list', 'fetched nothing!')



"""Extarct information from html source code."""


class Extractor(object):
    """Base class of Extractors."""
    @classmethod
    def create(cls, url: str):
        """return a specific Extractor depends on url."""
        for subclass in cls.__subclasses__():
            for pattern in subclass.pattern:
                if pattern.search(url):
                    return subclass()
        raise NotImplementedError



"""Library for Control flow."""
import asyncio
import platform
import time

from video_dl.args import Arguments
from video_dl.spider import Spider
from video_dl.toolbox import info
import video_dl.spiders  # pylint: disable=W0611
import video_dl.extractors  # pylint: disable=W0611


if platform.system() != 'Windows':
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def main():
    # get url from command line's augument and create a specifc spider.
    spider = Spider.create(Arguments().url)

    # start spider and download video.
    start_time = time.time()
    asyncio.run(spider.run())

    info('done', f'had wasted your time: {time.time() - start_time:.2f}s!')
